import numpy as np
import cv2

def apply_threshold(image, threshold_value):
    """
    Applies a binary threshold to the image.
    :param threshold_value: Threshold value for binarization.
    """
    _, binary_image = cv2.threshold(image, threshold_value, 255, cv2.THRESH_BINARY)
    return binary_image

def apply_gamma_correction(image, gamma):
    """
    Applies gamma correction to the image.
    :param gamma: Gamma value for correction.
    """
    gamma_corrected = np.array(255 * (image / 255) ** gamma, dtype='uint8')
    return gamma_corrected

def apply_negative(image):
    """
    Applies negative transformation to the image.
    """
    return 255 - image

def apply_log_transform(image, c=1):
    """
    Applies logarithmic transformation to the image.
    :param c: Constant for scaling the transformation.
    """
    log_transformed = c * np.log(1 + image)
    return np.array(log_transformed, dtype='uint8')

def apply_contrast_stretching(image, r_min=0, r_max=255):
    """
    Applies contrast stretching to the image.
    :param r_min: Minimum intensity value after stretching.
    :param r_max: Maximum intensity value after stretching.
    """
    stretched = np.interp(image, [np.min(image), np.max(image)], [r_min, r_max])
    return np.array(stretched, dtype='uint8')



import cv2
import numpy as np
import math  

def houghTransformCircle(image, method = cv2.HOUGH_GRADIENT_ALT, dp = 1, min_dist = 0,\
                    param1=200, param2=0.005, minRadius=0, maxRadius=0, kernelSize = 5):
    if min_dist == 0:
        min_dist = image.shape[0]//16
    # Convert to gray-scale
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    # # Apply Gaussian blur to reduce noise
    # img_blur = cv2.GaussianBlur(gray, (kernelSize,kernelSize), 0)

    # Blur the image to reduce noise
    img_blur = cv2.medianBlur(gray, kernelSize)
    # Apply hough transform on the image
    
    
    # gray: Input image (grayscale).
    # circles: A vector that stores sets of 3 values: xc,yc,r for each detected circle.
    # HOUGH_GRADIENT: Define the detection method. Currently this is the only one available in OpenCV.
    # dp = 1: The inverse ratio of resolution.
    # min_dist = gray.rows/16: Minimum distance between detected centers.
    # param_1 = 200: Upper threshold for the internal Canny edge detector.
    # param_2 = 100*: Threshold for center detection.
    # min_radius = 0: Minimum radius to be detected. If unknown, put zero as default.
    # max_radius = 0: Maximum radius to be detected. If unknown, put zero as default.
    # # Apply Hough transform to detect circles
    circles = cv2.HoughCircles(img_blur, method, dp, min_dist,
                            param1, param2, minRadius, maxRadius)
    
    # circles = cv2.HoughCircles(img_blur, cv2.HOUGH_GRADIENT, 1.3, 30, param1=150, param2=70, minRadius=0, maxRadius=0)

    # circles = cv2.HoughCircles(img_blur, cv2.HOUGH_GRADIENT, 1, image.shape[0]/64, \
    #     param1=200, param2=10, minRadius=0, maxRadius=30)
    
    # Convert the (x, y) coordinates and radius of the circles to integers
    circles = np.round(circles[0, :]).astype("int")

    return circles 


def houghTransformLines(image, cannyThresh1 = 50, cannyThresh2 = 200, cannyEdges =None,\
        canny_apertureSize = 3, canny_L2gradient =False,\
        HL_rho = 1, HL_theta = 0, HL_threshold = 150, HL_lines = None, HL_srn = 0, HL_stn = 0):  
    '''
    @brief Finds lines in a binary image using the standard Hough transform.

    The function implements the standard or standard multi-scale Hough transform algorithm for line detection. See <http://homepages.inf.ed.ac.uk/rbf/HIPR2/hough.htm> for a good explanation of Hough transform.
    
    @param cannyEdges	output edge map; single channels 8-bit image, which has the same size as image .
    @param cannyThresh1	first threshold for the hysteresis procedure.
    @param cannyThresh2	second threshold for the hysteresis procedure.
    @param canny_apertureSize	aperture size for the Sobel operator.
    @param canny_L2gradient	a flag, indicating whether a more accurate L_2 norm =\sqrt{(dI/dx)^2 + (dI/dy)^2} should be used to calculate the image gradient magnitude ( L2gradient=true ), or whether the default L_1 norm =|dI/dx|+|dI/dy| is enough ( L2gradient=false ).
    
    @param lines Output vector of lines. Each line is represented by a 2 or 3 element vector \f$(\rho, \theta)\f$ or \f$(\rho, \theta, \textrm{votes})\f$ . \f$\rho\f$ is the distance from the coordinate origin \f$(0,0)\f$ (top-left corner of the image). \f$\theta\f$ is the line rotation angle in radians ( \f$0 \sim \textrm{vertical line}, \pi/2 \sim \textrm{horizontal line}\f$ ). \f$\textrm{votes}\f$ is the value of accumulator.
    @param rho Distance resolution of the accumulator in pixels.
    @param theta Angle resolution of the accumulator in radians.
    @param threshold Accumulator threshold parameter. Only those lines are returned that get enough votes ( \f$>\texttt{threshold}\f$ ).
    @param srn For the multi-scale Hough transform, it is a divisor for the distance resolution rho . The coarse accumulator distance resolution is rho and the accurate accumulator resolution is rho/srn . If both srn=0 and stn=0 , the classical Hough transform is used. Otherwise, both these parameters should be positive.
    @param stn For the multi-scale Hough transform, it is a divisor for the distance resolution theta.
    @param min_theta For standard and multi-scale Hough transform, minimum angle to check for lines.
    Must fall between 0 and max_theta.
    @param max_theta For standard and multi-scale Hough transform, maximum angle to check for lines.
    Must fall between min_theta and CV_PI.

    The Hough Line Transform is a transform used to detect straight lines.
    To apply the Transform, first an edge detection pre-processing is desirable.
    
    Parameters
    ------------
    dst: Output of the edge detector. It should be a grayscale image (although in fact it is a binary one)
    lines: A vector that will store the parameters (r,θ) of the detected lines
    rho : The resolution of the parameter r in pixels. We use 1 pixel.
    theta: The resolution of the parameter θ in radians. We use 1 degree (CV_PI/180)
    threshold: The minimum number of intersections to "*detect*" a line
    srn and stn: Default parameters to zero. Check OpenCV reference for more info.
    '''      

            # Check if image is loaded fine
    if image is None:
        print('Error opening image!')
        return -1
    
    
    dst = cv2.Canny(image, cannyThresh1, cannyThresh2, cannyEdges, canny_apertureSize, canny_L2gradient)
    
    # Copy edges to the images that will display the results in BGR
    cdst = cv2.cvtColor(dst, cv2.COLOR_GRAY2BGR)
    
    if HL_theta == 0:
        HL_theta = np.pi / 180
    # lines = cv2.HoughLines(dst, 1, np.pi / 180, 150, None, 0, 0)
    lines = cv2.HoughLines(dst, rho= HL_rho, theta= HL_theta ,\
        threshold= HL_threshold, lines= HL_lines, srn=  HL_srn,stn= HL_stn)
    
    if lines is not None:
        for i in range(0, len(lines)):
            rho = lines[i][0][0]
            theta = lines[i][0][1]
            a = math.cos(theta)
            b = math.sin(theta)
            x0 = a * rho
            y0 = b * rho
            pt1 = (int(x0 + 1000*(-b)), int(y0 + 1000*(a)))
            pt2 = (int(x0 - 1000*(-b)), int(y0 - 1000*(a)))
            cv2.line(cdst, pt1, pt2, (0,0,255), 3, cv2.LINE_AA)
        
    return cdst, lines

# Probabilistic Line Transform
def houghTransformLinesP(image, cannyThresh1 = 50, cannyThresh2 = 200, cannyEdges =None,\
        canny_apertureSize = 3, canny_L2gradient =False,\
        HL_rho = 1, HL_theta = 0, HL_threshold = 150, HL_lines = None, HL_minLineLength = 50, HL_maxLineGap = 10):  
    '''
    HoughLinesP(image, rho, theta, threshold[, lines[, minLineLength[, maxLineGap]]]) -> canny, lines
    @brief Finds line segments in a binary image using the probabilistic Hough transform.
    The function implements the probabilistic Hough transform algorithm for line detection, described in @cite Matas00
    See the line detection example below:
    @include snippets/imgproc_HoughLinesP.cpp This is a sample picture the function parameters have been tuned for: image
    And this is the output of the above program in case of the probabilistic Hough transform: image
    
    
    @param cannyEdges	output edge map; single channels 8-bit image, which has the same size as image .
    @param cannyThresh1	first threshold for the hysteresis procedure.
    @param cannyThresh2	second threshold for the hysteresis procedure.
    @param canny_apertureSize	aperture size for the Sobel operator.
    @param canny_L2gradient	a flag, indicating whether a more accurate L_2 norm =\sqrt{(dI/dx)^2 + (dI/dy)^2} should be used to calculate the image gradient magnitude ( L2gradient=true ), or whether the default L_1 norm =|dI/dx|+|dI/dy| is enough ( L2gradient=false ).
    
    @param lines Output vector of lines. Each line is represented by a 4-element vector \f$(x_1, y_1, x_2, y_2)\f$ , where \f$(x_1,y_1)\f$ and \f$(x_2, y_2)\f$ are the ending points of each detected line segment.
    @param rho Distance resolution of the accumulator in pixels.
    @param theta Angle resolution of the accumulator in radians.
    @param threshold Accumulator threshold parameter. Only those lines are returned that get enough votes ( \f$>\texttt{threshold}\f$ ).
    @param minLineLength Minimum line length. Line segments shorter than that are rejected.
    @param maxLineGap Maximum allowed gap between points on the same line to link them.
    @sa LineSegmentDetector

    The Hough Line Transform is a transform used to detect straight lines.
    To apply the Transform, first an edge detection pre-processing is desirable.
    
    Parameters
    ------------
    dst: Output of the edge detector. It should be a grayscale image (although in fact it is a binary one)
    lines: A vector that will store the parameters (r,θ) of the detected lines
    rho : The resolution of the parameter r in pixels. We use 1 pixel.
    theta: The resolution of the parameter θ in radians. We use 1 degree (CV_PI/180)
    threshold: The minimum number of intersections to "*detect*" a line
    srn and stn: Default parameters to zero. Check OpenCV reference for more info.
    '''      

            # Check if image is loaded fine
    if image is None:
        print('Error opening image!')
        return -1
    

    dst = cv2.Canny(image, cannyThresh1, cannyThresh2, cannyEdges, canny_apertureSize, canny_L2gradient)
    
    # Copy edges to the images that will display the results in BGR
    cdst = cv2.cvtColor(dst, cv2.COLOR_GRAY2BGR)
    cdstP = np.copy(cdst)
    
    
    
    if HL_theta == 0:
        HL_theta = np.pi / 180
        
    linesP = cv2.HoughLinesP(dst, HL_rho, HL_theta, HL_threshold, HL_lines, HL_minLineLength, HL_maxLineGap)
    
    if linesP is not None:
        for i in range(0, len(linesP)):
            l = linesP[i][0]
            cv2.line(cdstP, (l[0], l[1]), (l[2], l[3]), (0,0,255), 3, cv2.LINE_AA)
        
    return cdstP, linesP


import numpy as np
import cv2
            

def add_gauss_noise(image, mean = 0, var = 0.1):
    '''
    Add Gaussian noise to the image.

    Parameters
    ----------
    mean : float, optional
        Mean of the Gaussian distribution to generate noise (default is 0).
    var : float, optional
        Variance of the Gaussian distribution to generate noise (default is 0.1).
    '''
    sigma = var**0.5
    
    
    gauss = np.random.normal(mean,sigma,image.size)
    gauss = gauss.reshape(image.shape[0],image.shape[1],image.shape[2]).astype('uint8')
    # Add the Gaussian noise to the image
    img_gauss = cv2.add(image,gauss)

    
    return img_gauss

def add_sp_noise(image, s_vs_p :float = 0.5, amount: float = 0.004):
    '''
    Add Salt & Pepper noise to the image.

    Parameters
    ----------
    s_vs_p : float, optional
        Ratio of salt to pepper (default is 0.5).
    amount : float, optional
        Overall proportion of image pixels to replace with noise (default is 0.004).
    '''
    out = np.copy(image)

    # Salt mode
    num_salt = np.ceil(amount * image.size * s_vs_p).astype(int)
    coords = tuple(map(lambda dim: np.random.randint(0, dim, num_salt), image.shape))
    out[coords] = 255

    # Pepper mode
    num_pepper = np.ceil(amount * image.size * (1. - s_vs_p)).astype(int)
    coords = tuple(map(lambda dim: np.random.randint(0, dim, num_pepper), image.shape))
    out[coords] = 0

    return out

def add_poisson_noise(image, **kwargs):
    '''
    Add Poisson noise to the image.

    The noise is added as per a Poisson distribution. This function does not take any additional parameters.
    '''
    # Convert the image to double data type
    image = image.astype(np.float64)

    # Scale the image to the range of 0-1
    image /= np.max(image)

    # Convert the image to represent counts in the range of 0-255
    image *= 255

    # Apply the Poisson noise
    noisy = np.random.poisson(image)

    # Normalize the noisy image
    noisy = noisy / np.max(noisy)
    
    noisy *= 255.
    noisy = noisy.astype(np.uint8) 

    return noisy



def add_speckle_noise(image, mean: float = 0, var: float = 0.5):
    """
    Add Speckle noise to the image.

    Speckle noise is a multiplicative noise. This function does not take any additional parameters.
    """
    
    sigma = var**0.5
    
    noise = np.random.normal(mean, sigma, image.shape)

    noise = noise.reshape(image.shape[0],image.shape[1],image.shape[2]).astype('uint8')
    noise = image + image * noise

    return noise




import numpy as np
import cv2
import matplotlib.pyplot as plt


def canny(image, min_val:int = 100, max_val:int = 200):
    imgShape = image.shape
    grayscale = image
    if len(imgShape) == 3:
        grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    edges = cv2.Canny(grayscale, min_val, max_val)
    
    return edges


def sobel(image, dx:int = 1, dy:int = 1, GaussianBlurKsize:int = 3):
    imgShape = image.shape
    grayscale = image
    if len(imgShape) == 3:
        grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        
    
    src = cv2.GaussianBlur(image, (GaussianBlurKsize, GaussianBlurKsize), 0)

    scale = 1
    delta = 0
    ddepth = cv2.CV_64F #cv2.CV_16S
    gray = src
    # gray = cv2.cvtColor(src, cv2.COLOR_BGR2GRAY)


    grad_x = cv2.Sobel(gray, ddepth, 1, 0, ksize=3, scale=scale, delta=delta, borderType=cv2.BORDER_DEFAULT)
    print('grad_x', grad_x.shape)
    # Gradient-Y
    # grad_y = cv.Scharr(gray,ddepth,0,1)
    grad_y = cv2.Sobel(gray, ddepth, 0, 1, ksize=3, scale=scale, delta=delta, borderType=cv2.BORDER_DEFAULT)


    abs_grad_x = cv2.convertScaleAbs(grad_x)
    abs_grad_y = cv2.convertScaleAbs(grad_y)


    edges = cv2.addWeighted(abs_grad_x, 0.5, abs_grad_y, 0.5, 0)
    
    return edges

def laplacian(image):
    imgShape = image.shape
    grayscale = image
    if len(imgShape) == 3:
        grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    # Apply Laplacian edge detection
    edges = cv2.Laplacian(grayscale, cv2.CV_8U)
    
    return edges

def prewitt(image):
    imgShape = image.shape
    grayscale = image
    if len(imgShape) == 3:
        grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    
    
    kernel_x = np.array([[1, 0, -1], [1, 0, -1], [1, 0, -1]])
    kernel_y = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]])
    edges_x = cv2.filter2D(grayscale, -1, kernel_x)
    edges_y = cv2.filter2D(grayscale, -1, kernel_y)
    edges = cv2.addWeighted(edges_x, 0.5, edges_y, 0.5, 0)
    
    return edges

def region_seg(image , thresholdFactor = 0.01):
    
    img = np.copy(image)
    imgShape = image.shape
    gray = image
    if len(imgShape) == 3:
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    # Apply thresholding to create a binary image
    ret, thresh = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY_INV+cv2.THRESH_OTSU)

    # Apply morphological operations to remove noise and fill holes
    kernel = np.ones((3,3), np.uint8)
    opening = cv2.morphologyEx(thresh, cv2.MORPH_OPEN, kernel, iterations=2)
    sure_bg = cv2.dilate(opening, kernel, iterations=3)
    dist_transform = cv2.distanceTransform(opening, cv2.DIST_L2, 5)
    ret, sure_fg = cv2.threshold(dist_transform, thresholdFactor*dist_transform.max(), 255, 0)
    sure_fg = np.uint8(sure_fg)
    unknown = cv2.subtract(sure_bg, sure_fg)

    # Apply the watershed algorithm to segment the image
    markers = cv2.connectedComponents(sure_fg)[1]
    markers += 1
    markers[unknown==255] = 0
    
    if img.shape[-1] != 3:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

    # Convert the image to the correct data type, if needed
    if img.dtype != np.uint8:
        img = img.astype(np.uint8)
    markers = cv2.watershed(img, markers)

    # Apply color mapping to visualize the segmentation
    colors = np.random.randint(0, 255, (len(np.unique(markers)), 3))
    color_map = np.zeros((img.shape[0], img.shape[1], 3), dtype=np.uint8)
    for i, c in enumerate(colors):
        color_map[markers == i+1] = c
        
    return color_map

def region_segmentation(image, seed_point, connectivity=4, Sobel_dx = 1, Sobel_dy = 1, Sobel_ksize = 5):
    '''
    
    does not work right
    '''
    
    # grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    elevation_map = canny(image) #cv2.Sobel(grayscale, cv2.CV_8U, Sobel_dx, Sobel_dy, ksize=Sobel_ksize)
    # elevation_map = sobel(image)
    coins = np.copy(image)

    # coin_shape = coins.shape

    # if len(coin_shape) == 3:
    #     coins = cv2.cvtColor(coins, cv2.COLOR_BGR2GRAY)
    
    coin_gray = coins
    if coins.shape[-1] == 3:
        coin_gray = cv2.cvtColor(coins, cv2.COLOR_BGR2GRAY)
    markers = np.zeros_like(coin_gray, np.uint8)

    
    markers[coin_gray < 30] = 1

    markers[coin_gray > 150] = 2

    print('elevation_map', elevation_map.shape)
    print('elevation_map', elevation_map.dtype)
    print('markers', markers.shape)
    print('markers', markers.dtype)
    
    if elevation_map.shape[-1] != 3:
        elevation_map = cv2.cvtColor(elevation_map, cv2.COLOR_GRAY2BGR)

    # Convert the image to the correct data type, if needed
    if elevation_map.dtype != np.uint8:
        elevation_map = elevation_map.astype(np.uint8)
        
        
    print('elevation_map:', elevation_map.dtype, elevation_map.shape)
    print('markers:', markers.dtype, markers.shape)
    segmentation = cv2.watershed(elevation_map, markers)


    h, w = segmentation.shape[:2]
    mask = np.zeros((h+2, w+2), np.uint8)
    segmentation = cv2.floodFill(segmentation , mask, (0,0), 255);#(segmentation - 1)
    # segmentation = ndi.binary_fill_holes(segmentation - 1)

    

    output = cv2.connectedComponentsWithStats(
        segmentation, coins, cv2.CV_32S)
    (numLabels, labeled_coins, stats, centroids) = output
    
    # labeled_coins, _ = ndi.label(segmentation)

    label_range = np.linspace(0, 1, 256)
    lut = np.uint8(plt.cm.viridis(label_range)[:,2::-1]*256).reshape(256, 1, 3) # replace viridis with a matplotlib colormap of your choice
    image_label_overlay = cv2.LUT(cv2.merge((labeled_coins, labeled_coins, labeled_coins)), lut)


    # image_label_overlay = label2rgb(labeled_coins, image=coins)
    return image_label_overlay




from skimage.feature import local_binary_pattern
from skimage.filters import gabor

def compute_lbp(image, radius:int = 1):
    n_points = 8 * radius
    lbp = local_binary_pattern(image, n_points, radius, method='uniform')
    return lbp

def apply_gabor_filters(image, frequency:float =0.2, theta: float =0.4):
    kernel_size = (31, 31)
    sigma = 5.0
    theta = 0.4
    lambd = 10.0
    gamma = 0.5
    gabor_filters = gabor(image, frequency=frequency, theta=theta)
    filtered_image = gabor_filters[0]
    return filtered_image



import cv2


def extract_hog(image, template):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
    
    hog = cv2.HOGDescriptor()
    features_image = hog.compute(image)
    features_templ = hog.compute(template)
    return features_image, features_templ

def extract_sift(image, template):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
    
    sift = cv2.SIFT_create()
    keypoints_image, descriptors_image = sift.detectAndCompute(image, None)
    keypoints_templ, descriptors_templ = sift.detectAndCompute(template, None)        
    return keypoints_image, descriptors_image, keypoints_templ, descriptors_templ

def extract_surf(image, template):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
    
    surf = cv2.SURF_create()
    keypoints_image, descriptors_image = surf.detectAndCompute(image, None)
    keypoints_templ, descriptors_templ = surf.detectAndCompute(template, None)
    return keypoints_image, descriptors_image, keypoints_templ, descriptors_templ

def extract_orb(image, template):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
    
    orb = cv2.ORB_create()
    keypoints_image, descriptors_image = orb.detectAndCompute(image, None)
    keypoints_templ, descriptors_templ = orb.detectAndCompute(template, None)
    return keypoints_image, descriptors_image, keypoints_templ, descriptors_templ

def extract_lbp(image, template):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
    
    radius = 1
    n_points = 8 * radius
    lbp = cv2.ximgproc.createLBPHFaceRecognizer(radius, n_points)
    _, features_image = lbp.compute(image)
    _, features_templ = lbp.compute(template)
    return features_image, features_templ

def match_template(image, template, method):
    
    image    = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
            
    # Create a brute-force matcher
    matcher = cv2.BFMatcher(cv2.NORM_L2) # for sift, surf, orb

    # Extract features from the image
    match method:
        case 'hog':
            f_i, f_t = extract_hog(image, template)
            # Perform template matching using HOG features                
            result = cv2.matchTemplate(f_i, f_t, cv2.TM_CCOEFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
    
            # Draw the best match on the image
            image_with_match = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            template_height, template_width = template.shape[:2]
            cv2.rectangle(image_with_match, max_loc, (max_loc[0] + template_width, max_loc[1] + template_height), (0, 0, 255), 2)
    
            return image_with_match
        case 'sift':
            k_i, d_i, k_t, d_t = extract_sift(image, template)
            # Match the descriptors
            matches = matcher.match(d_i, d_t)
    
            # Sort the matches by distance
            matches = sorted(matches, key=lambda x: x.distance)
    
            # Draw the best match on the image
            best_match = matches[0]
            image_with_match = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            cv2.rectangle(image_with_match, (int(k_i[best_match.queryIdx].pt[0]), int(k_i[best_match.queryIdx].pt[1])), (int(k_t[best_match.trainIdx].pt[0]), int(k_t[best_match.trainIdx].pt[1])), (0, 0, 255), 2)
            
            return image_with_match
            
        case 'surf':
            k_i, d_i, k_t, d_t = extract_surf(image, template)
            matches = matcher.match(d_i, d_t)
    
            # Sort the matches by distance
            matches = sorted(matches, key=lambda x: x.distance)
    
            # Draw the best match on the image
            best_match = matches[0]
            image_with_match = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            cv2.rectangle(image_with_match, (int(k_i[best_match.queryIdx].pt[0]), int(k_i[best_match.queryIdx].pt[1])), (int(k_t[best_match.trainIdx].pt[0]), int(k_t[best_match.trainIdx].pt[1])), (0, 0, 255), 2)
            
            return image_with_match
                            
        case 'orb':
            k_i, d_i, k_t, d_t = extract_orb(image, template)
            matches = matcher.match(d_i, d_t)
    
            # Sort the matches by distance
            matches = sorted(matches, key=lambda x: x.distance)
    
            # Draw the best match on the image
            best_match = matches[0]
            image_with_match = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            cv2.rectangle(image_with_match, (int(k_i[best_match.queryIdx].pt[0]), int(k_i[best_match.queryIdx].pt[1])), (int(k_t[best_match.trainIdx].pt[0]), int(k_t[best_match.trainIdx].pt[1])), (0, 0, 255), 2)
            
            return image_with_match
                            
            
        case 'lbp':
            f_i, f_t = extract_lbp(image, template)
            # Perform template matching using HOG features                
            result = cv2.matchTemplate(f_i, f_t, cv2.TM_CCOEFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
    
            # Draw the best match on the image
            image_with_match = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            template_height, template_width = template.shape[:2]
            cv2.rectangle(image_with_match, max_loc, (max_loc[0] + template_width, max_loc[1] + template_height), (0, 0, 255), 2)
    
            return image_with_match                                
        


import numpy as np
import cv2

def threshold_segmentation(image, threshold_value_1, threshold_value_2):
    grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    _, thresholded = cv2.threshold(grayscale, threshold_value_1, threshold_value_2, cv2.THRESH_BINARY)
    thresholded = np.where(thresholded > 0, 255, 0)                        
    return thresholded

def color_segmentation(image, lower_color, upper_color):
    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    lower_color = np.array(lower_color, np.uint8)
    upper_color = np.array(upper_color, np.uint8)
    mask = cv2.inRange(hsv, lower_color, upper_color)
    segmented = cv2.bitwise_and(image, image, mask=mask)
    return segmented

def watershed_segmentation(image):
    grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    _, thresholded = cv2.threshold(grayscale, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
    kernel = np.ones((3, 3), np.uint8)
    opening = cv2.morphologyEx(thresholded, cv2.MORPH_OPEN, kernel, iterations=2)
    sure_bg = cv2.dilate(opening, kernel, iterations=3)
    dist_transform = cv2.distanceTransform(opening, cv2.DIST_L2, 5)
    _, sure_fg = cv2.threshold(dist_transform, 0.7 * dist_transform.max(), 255, 0)
    sure_fg = np.uint8(sure_fg)
    unknown = cv2.subtract(sure_bg, sure_fg)
    _, markers = cv2.connectedComponents(sure_fg)
    markers += 1
    markers[unknown == 255] = 0
    segmented = cv2.watershed(image, markers)
    image[markers == -1] = [0, 0, 255]
    return segmented

def clustering_segmentation(image, num_clusters):
    reshaped = image.reshape(-1, 3)
    reshaped = np.float32(reshaped)
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.2)
    _, labels, centers = cv2.kmeans(reshaped, num_clusters, None, criteria, 10, cv2.KMEANS_RANDOM_CENTERS)
    centers = np.uint8(centers)
    segmented = centers[labels.flatten()]
    segmented = segmented.reshape(image.shape)
    return segmented



import cv2

def superResCUBIC(image, scale_factor:int = 3):
    
    # Upsample the image using bicubic interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_CUBIC)
    
    return img_hr


def superResAREA(image, scale_factor:int = 3):
    
    # Upsample the image using area interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_AREA)
    
    return img_hr


def superResBITS(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_BITS interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_BITS)
    
    return img_hr


def superResBITS2(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_BITS2 interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_BITS2)
    
    return img_hr


def superResLANCZOS4(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_LANCZOS4 interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_LANCZOS4)
    
    return img_hr


def superResLINEAR(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_LINEAR interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_LINEAR)
    
    return img_hr


def superResLINEAR_EXACT(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_LINEAR_EXACT interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_LINEAR_EXACT)
    
    return img_hr



def superResNEAREST(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_NEAREST interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_NEAREST)
    
    return img_hr

def superResNEAREST_EXACT(image, scale_factor:int = 3):
    
    # Upsample the image using INTER_NEAREST_EXACT interpolation
    img_hr = cv2.resize(image, None, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_NEAREST_EXACT)
    
    return img_hr





import cv2
from skimage.exposure import match_histograms

def apply_histogram_equalization(image):
    """
    Applies histogram equalization to the image.
    """   
    
    # تبدیل فضای رنگی از BGR به HSV
    hsv_img = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    # تطابق هیستوگرام فضای رنگی H
    hsv_img[:,:,0] = cv2.equalizeHist(hsv_img[:,:,0])

    # تبدیل فضای رنگی از HSV به BGR
    matched_img = cv2.cvtColor(hsv_img, cv2.COLOR_HSV2BGR)         
    return matched_img#cv2.equalizeHist(image)

def match_histogram(image, target_image):
    """
    Matches the histogram of the image to the histogram of the target image.
    :param target_image: Image whose histogram will be used as a reference.
    """        
    
    #  # تبدیل فضای رنگی از BGR به HSV
    # hsv_img = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    # hsv_target_img = cv2.cvtColor(target_image, cv2.COLOR_BGR2HSV)

    # # تطابق هیستوگرام فضای رنگی H
    # hsv_img[:,:,0] = cv2.equalizeHist(hsv_img[:,:,0], hsv_target_img[:,:,0])

    # # تبدیل فضای رنگی از HSV به BGR
    # matched_img = cv2.cvtColor(hsv_img, cv2.COLOR_HSV2BGR)
    
    #  # تبدیل فضای رنگی از BGR به HSV
    # hsv_img = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    # hsv_target_img = cv2.cvtColor(target_image, cv2.COLOR_BGR2HSV)

    # # تطابق هیستوگرام فضای رنگی H
    # hsv_img[:,:,0] = cv2.calcHist(hsv_img[:,:,0])
    # cv.normalize(hist_test2, hist_test2, alpha=0, beta=1, norm_type=cv.NORM_MINMAX)
    # hsv_target_img[:,:,0] = cv2.equalizeHist(hsv_target_img[:,:,0])

    # # تبدیل فضای رنگی از HSV به BGR
    # matched_img = cv2.cvtColor(hsv_img, cv2.COLOR_HSV2BGR)
    
    matched_img = match_histograms(image, target_image, channel_axis=0)
    
    return matched_img
        


import numpy as np
import cv2


def low_pass_filter(image, filter_size:int = 3):
    
    if len(image.shape) <3:
        ValueError("your image must be in RGB color space!")
    img = image
    new_img= np.zeros((img.shape[0], img.shape[1], 3))
    
    filter_=  np.ones((filter_size, filter_size))
    mod= filter_size**2
    
    for i in range(filter_.shape[0]//2, new_img.shape[0]-filter_.shape[0]//2):
        for j in range(filter_.shape[1]//2, new_img.shape[1]-filter_.shape[1]//2):
            
            x1= i-filter_.shape[0]//2
            x2= x1+ filter_.shape[0]
            y1= j- filter_.shape[1]//2
            y2= y1+ filter_.shape[1]
            
            arr0= img[ x1:x2 , y1:y2, 0]
            arr1= img[ x1:x2 , y1:y2, 1 ]
            arr2= img[ x1:x2 , y1:y2, 2 ]
            
            new_img[i][j][0]= int(np.dot(np.ravel(arr0), np.ravel(filter_).T )//mod)
            new_img[i][j][1]= int(np.dot(np.ravel(arr1), np.ravel(filter_).T )//mod)
            new_img[i][j][2]= int(np.dot(np.ravel(arr2), np.ravel(filter_).T )//mod)
    return np.array(new_img, dtype= 'uint8')

def high_pass_filter(image, filter_size:int = 3):
    img = image
    new_img= np.zeros((img.shape[0], img.shape[1], 3))
    
    filter_=  -1* np.ones((filter_size, filter_size))
    mod= filter_size**2
    filter_[filter_size//2][filter_size//2]= mod- 1
    
    
    for i in range(filter_.shape[0]//2, new_img.shape[0]-filter_.shape[0]//2):
        for j in range(filter_.shape[1]//2, new_img.shape[1]-filter_.shape[1]//2):
            
            x1= i-filter_.shape[0]//2
            x2= x1+ filter_.shape[0]
            y1= j- filter_.shape[1]//2
            y2= y1+ filter_.shape[1]
            
            arr0= img[ x1:x2 , y1:y2, 0]
            arr1= img[ x1:x2 , y1:y2, 1 ]
            arr2= img[ x1:x2 , y1:y2, 2 ]
            
            new_img[i][j][0]= int(np.dot(np.ravel(arr0), np.ravel(filter_).T )//mod)
            new_img[i][j][1]= int(np.dot(np.ravel(arr1), np.ravel(filter_).T )//mod)
            new_img[i][j][2]= int(np.dot(np.ravel(arr2), np.ravel(filter_).T )//mod)
    
            if new_img[i][j][0]<0:
                new_img[i][j][0]= 0
            elif new_img[i][j][0]> 255:
                new_img[i][j][0]= 255
            
            if new_img[i][j][1]<0:
                new_img[i][j][1]= 0
            elif new_img[i][j][1]> 255:
                new_img[i][j][1]= 255
            
            if new_img[i][j][2]<0:
                new_img[i][j][2]= 0
            elif new_img[i][j][2]> 255:
                new_img[i][j][2]= 255
                
    return np.array(new_img, dtype= 'uint8')

def high_boost_filter(image, filter_size:int = 3, const: int = 5):
    img = image
    new_img= np.zeros((img.shape[0], img.shape[1], 3))
    
    filter_=  -1*np.ones((filter_size, filter_size))
    mod= filter_size**2
    filter_[filter_size//2][filter_size//2]= (mod-1)*const
    
    
    for i in range(filter_.shape[0]//2, new_img.shape[0]-filter_.shape[0]//2):
        for j in range(filter_.shape[1]//2, new_img.shape[1]-filter_.shape[1]//2):
            
            x1= i-filter_.shape[0]//2
            x2= x1+ filter_.shape[0]
            y1= j- filter_.shape[1]//2
            y2= y1+ filter_.shape[1]
            
            arr0= img[ x1:x2 , y1:y2, 0]
            arr1= img[ x1:x2 , y1:y2, 1 ]
            arr2= img[ x1:x2 , y1:y2, 2 ]
            
            new_img[i][j][0]= np.dot(np.ravel(arr0), np.ravel(filter_).T )/mod
            new_img[i][j][1]= np.dot(np.ravel(arr1), np.ravel(filter_).T )/mod
            new_img[i][j][2]= np.dot(np.ravel(arr2), np.ravel(filter_).T )/mod
    
            if new_img[i][j][0]<0:
                new_img[i][j][0]= 0
            elif new_img[i][j][0]> 255:
                new_img[i][j][0]= 255
            
            if new_img[i][j][1]<0:
                new_img[i][j][1]= 0
            elif new_img[i][j][1]> 255:
                new_img[i][j][1]= 255
            
            if new_img[i][j][2]<0:
                new_img[i][j][2]= 0
            elif new_img[i][j][2]> 255:
                new_img[i][j][2]= 255
                
    return np.array(new_img, dtype= 'uint8')

def normal_val(x, y, sig):
    val= ((x**2)+ (y**2))/2.0
    val= val/(sig**2)
    fin_val= np.exp(-val)
    return fin_val


def gaussian_filter(image, filter_size = 3, variance = 0.1) :
    n= filter_size
    sig= np.sqrt(variance)
    
    img = image
    
    filter_= [np.array([ 0.0 for i in range(-(n//2), n//2 +1)]) for j in range(-(n//2), n//2 +1)]
    
    c= 1/normal_val(n//2, n//2, sig)
    for i in range(-(n//2), n//2 +1):
        for j in range(-(n//2), n//2 +1):
            filter_[i+ n//2][j+ n//2]= round(normal_val(i, j, sig)*c, 0)
    filter_= np.array(filter_)
    
    new_img= np.zeros((img.shape[0], img.shape[1], 3))
    mod= np.sum(filter_)
    for i in range(filter_.shape[0]//2, new_img.shape[0]-filter_.shape[0]//2):
        for j in range(filter_.shape[1]//2, new_img.shape[1]-filter_.shape[1]//2):

            x1= i-filter_.shape[0]//2
            x2= x1+ filter_.shape[0]
            y1= j- filter_.shape[1]//2
            y2= y1+ filter_.shape[1]

            arr0= img[ x1:x2 , y1:y2, 0]
            arr1= img[ x1:x2 , y1:y2, 1 ]
            arr2= img[ x1:x2 , y1:y2, 2 ]

            new_img[i][j][0]= int(np.dot(np.ravel(arr0), np.ravel(filter_).T )//mod)
            new_img[i][j][1]= int(np.dot(np.ravel(arr1), np.ravel(filter_).T )//mod)
            new_img[i][j][2]= int(np.dot(np.ravel(arr2), np.ravel(filter_).T )//mod)
    
    return np.array(new_img, dtype= 'uint8')


def mean_filter(image, filter_size:int = 3, padding=0):
    padded_image = cv2.copyMakeBorder(image, padding, padding, padding, padding, cv2.BORDER_CONSTANT)
    kernel_size = (filter_size, filter_size)
    return cv2.blur(padded_image, kernel_size)

def median_filter(image, filter_size:int = 3, padding=0):
    padded_image = cv2.copyMakeBorder(image, padding, padding, padding, padding, cv2.BORDER_CONSTANT)
    return cv2.medianBlur(padded_image, filter_size)

def adaptive_filter(image, filter_size = 3, padding=0, maxVal = 255, \
                    adaptiveMethod = cv2.ADAPTIVE_THRESH_GAUSSIAN_C, \
                    thresholdType = cv2.THRESH_BINARY, \
                    borderType = cv2.BORDER_CONSTANT, \
                    C = 1):
    
    imgB, imgG, imgR = cv2.split(image)
    padded_imageB = cv2.copyMakeBorder(imgB, padding, padding, padding, padding, borderType)
    f_b = cv2.adaptiveThreshold(padded_imageB, maxVal, adaptiveMethod, thresholdType, filter_size, C)
    
    padded_imageG = cv2.copyMakeBorder(imgG, padding, padding, padding, padding, borderType)
    f_g = cv2.adaptiveThreshold(padded_imageG, maxVal, adaptiveMethod, thresholdType, filter_size, C)
    
    padded_imageR = cv2.copyMakeBorder(imgR, padding, padding, padding, padding, borderType)
    f_r = cv2.adaptiveThreshold(padded_imageR, maxVal, adaptiveMethod, thresholdType, filter_size, C)
    
    return cv2.merge([f_b, f_g, f_r])
    # padded_image = cv2.copyMakeBorder(image, padding, padding, padding, padding, cv2.BORDER_CONSTANT)
    # return cv2.adaptiveThreshold(padded_image, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, filter_size, 2)




import numpy as np
import cv2

def fftshiftRGB(image):
    '''
    This method:
    1. Convert BGR image to RGB 
    2. Then compute the nD fft of the image using NumPy's fftn function
    3. Then Shift the zero-frequency component to the center 
    4. at last return shifted component in RGB format
    
    Usage
    --------
    For using the output, you must write this codes and then show the image:
    magnitude_spectrum = visibleFactory * np.log(np.abs(fshiftRGB)).astype(np.uint8)
    magnitude_spectrum =  cv2.cvtColor(magnitude_spectrum.astype(np.uint8), cv2.COLOR_RGB2BGR)  
    
    When we compute the magnitude spectrum using NumPy's abs and log functions, 
    and scale it by a factor of visibleFactory (like 20) to make it more visible. 
    '''
    img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    f = np.fft.fftn(img)

    fshiftRGB = np.fft.fftshift(f)
    
    return fshiftRGB

def ifftshiftBGR(fshift):
    f_ishift = np.fft.ifftshift(fshift)
    img_back = np.fft.ifftn(f_ishift)
    img_back = np.abs(img_back)
    return img_back

def fftshift(image):
    imgB, imgG, imgR = cv2.split(image)
    
    fb = np.fft.fft2(imgB)
    fshiftB = np.fft.fftshift(fb)
            
    fg = np.fft.fft2(imgG)
    fshiftG = np.fft.fftshift(fg)
    
    fr = np.fft.fft2(imgR)
    fshiftR = np.fft.fftshift(fr)
    
    return fshiftB, fshiftG, fshiftR



def ifftshift(fshiftB, fshiftG, fshiftR):
    f_ishiftB = np.fft.ifftshift(fshiftB)
    img_backB = np.fft.ifftn(f_ishiftB)
    img_backB = np.abs(img_backB)
    
    f_ishiftG = np.fft.ifftshift(fshiftG)
    img_backG = np.fft.ifftn(f_ishiftG)
    img_backG = np.abs(img_backG)
    
    f_ishiftR = np.fft.ifftshift(fshiftR)
    img_backR = np.fft.ifftn(f_ishiftR)
    img_backR = np.abs(img_backR)
    
    img_back = cv2.merge([img_backB, img_backG, img_backR]).astype(int)
    return img_back

def create_mask(image, radius):
    base = np.zeros(image.shape[:2])
    cv2.circle(base, (image.shape[1]//2, image.shape[0]//2), int(radius), (1, 1, 1), -1, 8, 0)
    return base

def low_pass_filter(image, radius):
    fshiftB, fshiftG, fshiftR = fftshift(image=image)
    
    mask = create_mask(radius)
    
    fshiftB = fshiftB * mask
    
    fshiftG = fshiftG * mask
    
    fshiftR = fshiftR * mask
    ifshift = ifftshift(fshiftB, fshiftG, fshiftR)
    ifshift = cv2.normalize(ifshift, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
    
    return ifshift


def high_pass_filter(image, radius):
    
    fshiftB, fshiftG, fshiftR = fftshift(image)
    
    mask = create_mask(radius)
    
    fshiftB = fshiftB * (1 - mask)
    fshiftG = fshiftG * (1 - mask)
    fshiftR = fshiftR * (1 - mask)
    ifshift = ifftshift(fshiftB, fshiftG, fshiftR)
    ifshift = cv2.normalize(ifshift, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
    
    return ifshift
    
    

def band_pass_filter(image, min_radius, max_radius):
    
    
    
    fshiftB, fshiftG, fshiftR = fftshift(image)
    
    min_mask = create_mask(min_radius)
    max_mask = create_mask(max_radius)
    band_mask = max_mask - min_mask
    
    fshiftB = fshiftB * band_mask
    fshiftG = fshiftG * band_mask
    fshiftR = fshiftR * band_mask
    ifshift = ifftshift(fshiftB, fshiftR, fshiftR)
    ifshift = cv2.normalize(ifshift, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
    
    return ifshift
    
    

def band_reject_filter(image, min_radius, max_radius):
    
    fshiftB, fshiftG, fshiftR = fftshift(image)
    
    min_mask = create_mask(min_radius)
    max_mask = create_mask(max_radius)
    band_mask = max_mask - min_mask
    
    fshiftB = np.multiply(fshiftB , (1 - band_mask))
    fshiftG = np.multiply(fshiftG , (1 - band_mask))
    fshiftR = np.multiply(fshiftR , (1 - band_mask))
    ifshift = ifftshift(fshiftB, fshiftG, fshiftR)
    ifshift = cv2.normalize(ifshift, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
    
    return ifshift

def wiener_filtering(image, h, K):
    '''
    :param input_signal: 
    :param h: 
    :param K:
    :return:
    '''
    output_signal = [] 

    output_signal_fft = []

    input_signal_cp = np.copy(image) 

    input_signal_cp_fft = np.fft.fft2(input_signal_cp) 

    h_fft = np.fft.fft2(h) 

    h_abs_square = np.abs(h_fft)**2 

    output_signal_fft = np.conj(h_fft) / (h_abs_square + K)

    output_image = np.abs(np.fft.ifft2(output_signal_fft * input_signal_cp_fft)) 

    return output_image



import cv2

import heapq
import os
from collections import defaultdict

def compress_using_jpeg(image, quality):
    """
    Compresses the image using JPEG compression.
    :param quality: Compression quality (0-100), higher value means better quality.
    """
    encoded_image, compressed_image = cv2.imencode('.jpg', image, [int(cv2.IMWRITE_JPEG_QUALITY), quality])
    return compressed_image

def compress_using_png(image, compression_level):
    """
    Compresses the image using PNG compression.
    :param compression_level: Compression level (0-9), higher value means higher compression.
    """
    encoded_image, compressed_image = cv2.imencode('.png', image, [int(cv2.IMWRITE_PNG_COMPRESSION), compression_level])
    return compressed_image
def compress_using_huffman(IMAGE_PATH):
    
    h = HuffmanCoding(IMAGE_PATH)
    output_path = h.compress()
    return output_path
        
            

class HuffmanCoding:
    
    def __init__(self, image_path):
        self.image_path = image_path
        self.heap = []
        self.codes = {}
        self.reverse_mapping = {}
    
    class HeapNode:
        
        def __init__(self, pixel_val, freq):
            self.pixel_val = pixel_val
            self.freq = freq
            self.left = None
            self.right = None
        
        def __lt__(self, other):
            return self.freq < other.freq
        
        def __eq__(self, other):
            if(other == None):
                return False
            if(not isinstance(other, self)):
                return False
            return self.freq == other.freq
    
    def make_frequency_dict(self, pixels):
        frequency = defaultdict(int)
        for pixel in pixels:
            frequency[pixel] += 1
        return frequency
    
    def make_heap(self, frequency):
        for key in frequency:
            node = self.HeapNode(key, frequency[key])
            heapq.heappush(self.heap, node)
    
    def merge_nodes(self):
        while(len(self.heap)>1):
            node1 = heapq.heappop(self.heap)
            node2 = heapq.heappop(self.heap)
            merged = self.HeapNode(None, node1.freq + node2.freq)
            merged.left = node1
            merged.right = node2
            heapq.heappush(self.heap, merged)
    
    def make_codes_helper(self, root, current_code):
        if(root == None):
            return
        if(root.pixel_val != None):
            self.codes[root.pixel_val] = current_code
            self.reverse_mapping[current_code] = root.pixel_val
            return
        self.make_codes_helper(root.left, current_code + "0")
        self.make_codes_helper(root.right, current_code + "1")
    
    def make_codes(self):
        root = heapq.heappop(self.heap)
        current_code = ""
        self.make_codes_helper(root, current_code)
    
    def get_encoded_image(self, pixels):
        encoded_image = ""
        for pixel in pixels:
            encoded_image += self.codes[pixel]
        return encoded_image
    
    def pad_encoded_image(self, encoded_image):
        extra_padding = 8 - len(encoded_image) % 8
        for i in range(extra_padding):
            encoded_image += "0"
        padded_info = "{0:08b}".format(extra_padding)
        padded_encoded_image = padded_info + encoded_image
        return padded_encoded_image
    
    def get_byte_array(self, padded_encoded_image):
        if(len(padded_encoded_image) % 8 != 0):
            print("Encoded image not padded properly")
            exit(0)
        b = bytearray()
        for i in range(0, len(padded_encoded_image), 8):
            byte = padded_encoded_image[i:i+8]
            b.append(int(byte, 2))
        return b
    
    def compress(self):
        filename, ext = os.path.splitext(self.image_path)
        output_path = filename + ".bin"
        with open(self.image_path, 'rb') as image_file, open(output_path, 'wb') as output:
            pixels = []
            byte = image_file.read(1)
            while(byte != b''):
                pixels.append(int.from_bytes(byte, byteorder='big'))
                byte = image_file.read(1)
            frequency = self.make_frequency_dict(pixels)
            self.make_heap(frequency)
            self.merge_nodes()
            self.make_codes()
            encoded_image = self.get_encoded_image(pixels)
            padded_encoded_image = self.pad_encoded_image(encoded_image)
            b = self.get_byte_array(padded_encoded_image)
            output.write(bytes(b))
        print("Compressed successfully!")
        return output_path
                        
            